# coding=utf-8
"""
Celery tasks related to document transformation and preview.
"""
from __future__ import absolute_import

import logging
from contextlib import contextmanager

from celery import shared_task

from abilian.core.extensions import db
from abilian.services import converter, get_service
from abilian.services.conversion import ConversionError

logger = logging.getLogger(__package__)


@contextmanager
def get_document(document_id, session=None):
    """ Context manager that yields (session, document).
    """
    from .models import Document

    doc_session = session
    if session is None:
        doc_session = db.create_scoped_session()

    with doc_session.begin_nested():
        query = doc_session.query(Document)
        document = query.get(document_id)
        yield (doc_session, document)

    # cleanup
    if session is None:
        doc_session.commit()
        doc_session.close()


@shared_task
def process_document(document_id):
    """ Run document processing chain.
    """
    with get_document(document_id) as (session, document):
        if document is None:
            return

        # True = Ok, None means no check performed (no antivirus present)
        is_clean = _run_antivirus(document)
        if is_clean is False:
            return

    preview_document.delay(document_id)
    convert_document_content.delay(document_id)


def _run_antivirus(document):
    antivirus = get_service('antivirus')
    if antivirus and antivirus.running:
        is_clean = antivirus.scan(document.content_blob)
        if 'antivirus_task' in document.content_blob.meta:
            del document.content_blob.meta['antivirus_task']
        return is_clean
    return None


@shared_task
def antivirus_scan(document_id):
    """ Return antivirus.scan() result
    """
    with get_document(document_id) as (session, document):
        if document is None:
            return
        return _run_antivirus(document)


@shared_task
def preview_document(document_id):
    """ Compute the document preview images with its default preview size.
    """
    with get_document(document_id) as (session, document):
        if document is None:
            # deleted after task queued, but before task run
            return

        try:
            converter.to_image(document.content_digest, document.content,
                               document.content_type, 0, document.preview_size)
        except ConversionError as e:
            logger.info('Preview failed: %s',
                        str(e),
                        exc_info=True,
                        extra={'stack': True})


@shared_task
def convert_document_content(document_id):
    """ Convert document content.
    """
    with get_document(document_id) as (session, document):
        if document is None:
            # deleted after task queued, but before task run
            return

        error_kwargs = dict(exc_info=True, extra={'stack': True})

        conversion_args = (document.content_digest, document.content,
                           document.content_type)

        if document.content_type == "application/pdf":
            document.pdf = document.content
        else:
            try:
                document.pdf = converter.to_pdf(*conversion_args)
            except ConversionError as e:
                document.pdf = ""
                logger.info("Conversion to PDF failed: %s", str(e),
                            **error_kwargs)

        try:
            document.text = converter.to_text(document.content_digest,
                                              document.content,
                                              document.content_type)
        except ConversionError as e:
            document.text = u""
            logger.info("Conversion to text failed: %s", str(e), **error_kwargs)

        document.extra_metadata = {}
        try:
            document.extra_metadata = converter.get_metadata(*conversion_args)
        except ConversionError as e:
            logger.warning("Metadata extraction failed: %s", str(e),
                           **error_kwargs)
        except UnicodeDecodeError as e:
            logger.error("Unicode issue: %s", str(e), **error_kwargs)
        except Exception as e:
            logger.error("Other issue: %s", str(e), **error_kwargs)

        if document.text:
            import langid
            document.language = langid.classify(document.text)[0]

        document.page_num = document.extra_metadata.get("PDF:Pages", 1)



# -*- coding: utf-8 -*-
"""
Document management blueprint.
"""
from __future__ import absolute_import

from flask import g

from abilian.i18n import _l
from abilian.sbe.apps.communities.blueprint import Blueprint
from abilian.web import nav

from ..actions import register_actions

__all__ = ['documents']

documents = Blueprint("documents",
                      __name__,
                      url_prefix="/docs",
                      template_folder="../templates")
route = documents.route
documents.record_once(register_actions)


@documents.url_value_preprocessor
def init_document_values(endpoint, values):
    g.current_tab = 'documents'

    g.breadcrumb.append(
        nav.BreadcrumbItem(label=_l(u'Documents'),
                           url=nav.Endpoint('documents.index',
                                            community_id=g.community.slug)))



"""
The blueprint for this app.
"""

from flask import Blueprint, redirect, render_template, url_for

from abilian.core.extensions import db
from abilian.core.util import get_params
from abilian.sbe.apps.wall.presenters import ActivityEntryPresenter
from abilian.sbe.apps.wall.util import get_recent_entries

from ..models import Message

__all__ = ['social']

MAX_LAST_USERS = 15
MAX_MESSAGES = 15

social = Blueprint("social",
                   __name__,
                   url_prefix='/social',
                   template_folder='../templates')
route = social.route


@route("/")
def home():
    ctx = {}
    entries = get_recent_entries(num=50)
    entries = ActivityEntryPresenter.wrap_collection(entries)
    ctx['activity_entries'] = entries

    return render_template("social/home.html", **ctx)


@route("/stream/<stream_name>")
def stream(stream_name):
    pass


@route("/", methods=['POST'])
def share():
    # TODO: better error control / feedback
    d = get_params(Message.__editable__)
    if not d.get('content'):
        return redirect(url_for(".home"))

    message = Message(**d)
    db.session.add(message)

    #tags = message.tags
    #values = [ {'tag': tag, 'message_id': message.id} for tag in tags ]
    #db.engine.execute(tagging.insert(), values)

    db.session.commit()
    return redirect(url_for(".home"))



# coding=utf-8
"""
"""
from __future__ import absolute_import

from datetime import datetime
from urllib import quote

import sqlalchemy as sa
from flask import (current_app, flash, g, make_response, redirect,
                   render_template, request)
from flask_login import current_user
from flask_mail import Message
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.extensions import db, mail
from abilian.core.signals import activity
from abilian.i18n import _, render_template_i18n
from abilian.sbe.apps.communities.views import default_view_kw
from abilian.services import audit_service
from abilian.services.conversion import converter
from abilian.services.image import FIT, resize
from abilian.web import csrf, url_for
from abilian.web.action import actions
from abilian.web.frontend import add_to_recent_items
from abilian.web.views import default_view

from ..models import Document
from ..repository import repository
from ..tasks import convert_document_content, preview_document
from .util import (breadcrumbs_for, check_manage_access, check_read_access,
                   check_write_access, edit_object, get_document, match)
from .views import documents

route = documents.route

MAX_PREVIEW_SIZE = 1000


@default_view(documents, Document, id_attr='doc_id', kw_func=default_view_kw)
@route("/doc/<int:doc_id>")
def document_view(doc_id):
    doc = get_document(doc_id)
    check_read_access(doc)
    doc.ensure_antivirus_scheduled()
    db.session.commit()

    bc = breadcrumbs_for(doc)
    actions.context['object'] = doc

    if doc.content_type.startswith("image/"):
        add_to_recent_items(doc, "image")
    else:
        add_to_recent_items(doc, "document")

    has_preview = doc.has_preview()
    audit_entries = audit_service.entries_for(doc)

    ctx = dict(doc=doc,
               audit_entries=audit_entries,
               breadcrumbs=bc,
               folder=doc.parent,
               has_preview=has_preview,
               csrf_token=csrf.field())
    return render_template("documents/document.html", **ctx)


#
# Actions on documents
#
@route("/doc/<int:doc_id>", methods=['POST'])
@csrf.protect
def document_edit(doc_id):
    doc = get_document(doc_id)
    check_write_access(doc)

    changed = edit_object(doc)

    if changed:
        db.session.commit()
        flash(_(u"Document properties successfully edited."), "success")
    else:
        flash(_(u"You didn't change any property."), "success")

    return redirect(url_for(doc))


@route("/doc/<int:doc_id>/delete", methods=['POST'])
@csrf.protect
def document_delete(doc_id):
    doc = get_document(doc_id)
    check_write_access(doc)

    parent_folder = doc.parent
    repository.delete_object(doc)
    db.session.commit()

    flash(_(u"File successfully deleted."), "success")
    return redirect(url_for(parent_folder))


@route("/doc/<int:doc_id>/upload", methods=['POST'])
@csrf.protect
def document_upload(doc_id):
    doc = get_document(doc_id)
    check_write_access(doc)

    fd = request.files['file']
    doc.set_content(fd.read(), fd.content_type)
    del doc.lock

    self = current_app._get_current_object()
    activity.send(self, actor=g.user, verb="update", object=doc)
    db.session.commit()
    flash(_(u"New version successfully uploaded"), "success")
    return redirect(url_for(doc))


@route("/doc/<int:doc_id>/download")
def document_download(doc_id, attach=None):
    """Download the file content."""
    doc = get_document(doc_id)

    response = make_response(doc.content)
    response.headers['content-length'] = doc.content_length
    response.headers['content-type'] = doc.content_type

    if attach is None:
        attach = request.args.get('attach')

    if (attach or not match(doc.content_type,
                            ("text/plain", "application/pdf", "image/*"))):
        # Note: we omit text/html for security reasons.
        quoted_filename = quote(doc.title.encode('utf8'))
        response.headers['content-disposition'] \
          = 'attachment;filename="{}"'.format(quoted_filename)

    return response


@route('/doc/<int:doc_id>/checkin_checkout', methods=['POST'])
def checkin_checkout(doc_id):
    doc = get_document(doc_id)
    action = request.form.get('action')

    if action not in (u'checkout', u'lock', u'unlock'):
        raise BadRequest(u'Unknown action: %r' % action)

    session = sa.orm.object_session(doc)

    if action in (u'lock', u'checkout'):
        doc.lock = current_user
        d = doc.updated_at
        # prevent change of last modification date
        doc.updated_at = datetime.utcnow()
        session.flush()
        doc.updated_at = d
        session.commit()

        if action == u'lock':
            return redirect(url_for(doc))
        elif action == u'checkout':
            return document_download(doc_id, attach=True)

    if action == u'unlock':
        del doc.lock
        d = doc.updated_at
        # prevent change of last modification date
        doc.updated_at = datetime.utcnow()
        session.flush()
        doc.updated_at = d
        session.commit()
        return redirect(url_for(doc))


def preview_missing_image():
    response = redirect(url_for('abilian_sbe_static',
                                filename='images/preview_missing.png'))
    response.headers['Cache-Control'] = 'no-cache'
    return response


@route("/doc/<int:doc_id>/preview_image")
def document_preview_image(doc_id):
    """Returns a preview (image) for the file given by its id."""

    doc = get_document(doc_id)

    if not doc.antivirus_ok:
        return preview_missing_image()

    size = int(request.args.get("size", 0))

    # Just in case
    if size > MAX_PREVIEW_SIZE:
        size = MAX_PREVIEW_SIZE

    # compute image if size != standard document size
    get_image = (converter.get_image if size == doc.preview_size else
                 converter.to_image)

    content_type = "image/jpeg"

    if doc.content_type.startswith('image/svg'):
        image = doc.content
        content_type = doc.content_type
    elif doc.content_type.startswith("image/"):
        image = doc.content
        if size:
            image = resize(image, size, size, mode=FIT)
    else:
        page = int(request.args.get("page", 0))
        try:
            image = get_image(doc.digest, doc.content, doc.content_type, page,
                              size)
        except:
            # TODO: use generic "conversion failed" image
            image = ""

    if not image:
        return preview_missing_image()

    response = make_response(image)
    response.headers['content-type'] = content_type
    return response


@route("/doc/<int:doc_id>/refresh_preview")
def refresh_preview(doc_id):
    """ Force to compute a new preview. """
    doc = get_document(doc_id)
    if not doc:
        raise NotFound()

    ct = doc.find_content_type(doc.content_type)
    if ct != doc.content_type:
        doc.content_type = ct
        db.session.commit()

    check_manage_access(doc)
    convert_document_content.apply([doc_id])
    preview_document.apply([doc_id])
    return redirect(url_for(doc))


@route("/doc/<int:doc_id>/send", methods=['POST'])
@csrf.protect
def document_send(doc_id):
    doc = get_document(doc_id)

    recipient = request.form.get("recipient")
    user_msg = request.form.get('message')

    site_name = u'[{}] '.format(current_app.config['SITE_NAME'])
    sender_name = g.user.name
    subject = site_name + _(u'{sender} sent you a file').format(sender=sender_name)
    msg = Message(subject)
    msg.sender = g.user.email
    msg.recipients = [recipient]
    msg.body = render_template_i18n('documents/mail_file_sent.txt',
                                    sender_name=sender_name,
                                    message=user_msg,
                                    document_url=url_for(doc),
                                    filename=doc.title)

    filename = doc.title
    msg.attach(filename, doc.content_type, doc.content)

    mail.send(msg)
    flash(_(u"Email successfully sent"), "success")

    return redirect(url_for(doc))


@route("/doc/<int:doc_id>/preview")
def document_preview(doc_id):
    doc = get_document(doc_id)
    if not doc.antivirus_ok:
        return "Waiting for antivirus to finish"

    if doc.content_type == "application/pdf":
        return redirect(url_for(".document_view_pdf",
                                community_id=g.community.slug,
                                doc_id=doc.id))

    else:
        return redirect(url_for(".document_download",
                                community_id=g.community.slug,
                                doc_id=doc.id))


@route("/doc/<int:doc_id>/view_pdf")
def document_view_pdf(doc_id):
    doc = get_document(doc_id)
    if not doc.antivirus_ok:
        return "Waiting for antivirus to finish"

    return render_template("documents/view_pdf.html",
                           pdf_url=url_for(".document_download",
                                           community_id=g.community.slug,
                                           doc_id=doc.id))

#
# Tagging (currently not used!)
#
#@route("/tag")
#def tag():
#  tag = request.args.get("tag")
#  if not tag:
#    return redirect("/dm/")
#
#  bc = [dict(path="/", label="Home"), dict(path="/dm/", label="DM")]
#  bc += [dict(path=request.path, label="Filter by tag")]
#  # TODO ...
#  docs = Document.query.filter(Document.tags.like("%" + tag + "%"))
#  docs = list(docs.all())
#  docs = [f for f in docs if tag in f.tags.split(",")]
#  title = "Files filtered by tag: %s" % tag
#  return render_template("dm/home.html", title=title, breadcrumbs=bc,
# files=docs)
#
#
#@route("/<int:file_id>/tag", methods=['POST'])
#def tag_post(file_id):
#  doc = get_document(file_id)
#  tags = request.form.get("tags")
#
#  doc.tags = tags
#  self = current_app._get_current_object()
#  activity.send(self, actor=g.user, verb="tag", object=doc)
#
#  db.session.commit()
#
#  flash("Tags successfully successfully updated", "success")
#  return redirect(url_for(".document_view", doc_id=doc.id))
